import json

from django.db import models
from django import forms
from django.template.defaultfilters import slugify
from django.utils.datastructures import SortedDict
from django.utils import simplejson as json

from .exceptions import (ChoiceDoesNotExist, ChoiceAlreadyExists,
                         FieldDoesNotExist, FieldAlreadyExists,
                         WysiwygFormsException)
from .utils import field_type_has_choices

class Form(models.Model):
    slug        = models.SlugField(editable=False)
    name        = models.CharField(max_length=250)
    description = models.TextField()

    # Don't use Django's serializers because it is just too cluttered.
    class JSONEncoder(json.JSONEncoder):
        def default(self, obj):
            if isinstance(obj, Form):
                return { "name"        : obj.name,
                         "id"          : obj.id,
                         "description" : obj.description,
                         "fields"      : [{ "label"     : f.label,
                                            "name"      : f.slug,
                                            "type"      : f.type,
                                            "widget"    : f.widget,
                                            "help_text" : f.help_text,
                                            "required"  : f.required,
                                            "position"  : f.position,
                                            "choices"   : [{ "label"    : c.label,
                                                             "position" : c.position }
                                                           for c in f.choices] }
                                          for f in obj.fields] }
            else:
                return super(JSONEncoder, self).default(obj)

    def __init__(self, *args, **kwargs):
        super(Form, self).__init__(*args, **kwargs)
        self._fields = None

    def _ensure_field_positions(self):
        """
        Iterates through self.fields and makes sure that they have the correct
        position attribute.
        """
        for f, i in zip(self.fields, range(len(self.fields))):
            f.position = i

    def save(self, *args, **kwargs):
        self.slug = slugify(self.name).replace("-", "_")[:50]
        for field in self.fields:
            field.save()
        super(Form, self).save(*args, **kwargs)

    def as_django_form(self):
        properties = SortedDict()
        for field in self.fields:
            properties[field.slug] = field.as_django_form_field()
        return type(str(self.slug), (forms.Form,), properties)

    def as_json(self):
        return json.dumps(self, cls=self.JSONEncoder)

    @property
    def fields(self):
        if self._fields is None:
            self._fields = list(self._field_set.all().order_by("position"))
        return self._fields

    def get_field(self, label):
        """
        Returns the field whose label is `label`, otherwise throws
        `FieldDoesNotExist`.
        """
        try:
            return (f for f in self.fields if f.label == label).next()
        except StopIteration:
            raise FieldDoesNotExist(
                "Tried to find the field '%s' but it doesn't exist." % label
                )

    def add_field(self, field_label, **field_properties):
        if any(f for f in self.fields if f.label == field_label):
            raise FieldAlreadyExists(
                "Tried to add field '%s' but it already is a field." % field_label
                )
        else:
            position = len(self.fields)
            field = Field.objects.create(form=self,
                                         label=field_label,
                                         position=position,
                                         **field_properties)
            self.fields.append(field)
            self._ensure_field_positions()
            return field

    def remove_field(self, field_label):
        field = self.get_field(field_label)
        self.fields.remove(field)
        self._ensure_field_positions()
        field.delete()
        return field

    def move_field(self, label, new_index):
        field = self.get_field(label)
        old_index = self.fields.index(field)
        self.fields.insert(new_index, self.fields.pop(old_index))
        self._ensure_field_positions()
        return field

class Field(models.Model):
    form      = models.ForeignKey(Form, related_name="_field_set")
    slug      = models.SlugField(editable=False)
    label     = models.CharField(max_length=250)
    help_text = models.CharField(max_length=250, default="")
    type      = models.CharField(max_length=250, default="CharField")
    position  = models.IntegerField(editable=False)
    required  = models.BooleanField(default=True)
    widget    = models.CharField(max_length=250, default="TextInput")

    def __init__(self, *args, **kwargs):
        super(Field, self).__init__(*args, **kwargs)
        self._choices = None

    def _ensure_choice_positions(self):
        """
        Iterates through self.choices and makes sure that they have the correct
        position attribute.
        """
        for c, i in zip(self.choices, range(len(self.choices))):
            c.position = i

    def save(self, *args, **kwargs):
        self.slug = slugify(self.label).replace("-", "_")[:50]
        for choice in self.choices:
            choice.save()
        super(Field, self).save(*args, **kwargs)

    def as_django_form_field(self):
        # TODO: catch exceptions and display helpful error messages
        field_properties = { "help_text": self.help_text,
                             "required": self.required }
        if field_type_has_choices(self.type) and self.choices:
            field_properties["choices"] = ((c.slug, c.label)
                                           for c in self.choices)
        if self.widget:
            field_properties["widget"] = getattr(forms.widgets, self.widget)()
        return getattr(forms.fields, self.type)(**field_properties)

    @property
    def choices(self):
        if self._choices is None:
            self._choices = list(self._choice_set.all().order_by("position"))
        return self._choices

    def get_choice(self, label):
        """
        Returns the choice whose label is `label`, otherwise throws
        `ChoiceDoesNotExist`.
        """
        try:
            return (c for c in self.choices if c.label == label).next()
        except StopIteration:
            raise ChoiceDoesNotExist(
                "Tried to find the choice '%s' but it doesn't exist." % label
                )

    def add_choice(self, choice_label):
        if not field_type_has_choices(self.type):
            raise WysiwygFormsException(
                "The field type '%s' doesn't support choices" % self.type
                )
        else:
            if any(c for c in self.choices if c.label == choice_label):
                raise ChoiceAlreadyExists(
                    "Tried to add choice '%s' but it already is a choice." % choice_label
                    )
            else:
                position = len(self.choices)
                choice = Choice.objects.create(field=self,
                                               label=choice_label,
                                               position=position)
                self.choices.append(choice)
                self._ensure_choice_positions()
                return choice

    def remove_choice(self, choice_label):
        choice = self.get_choice(choice_label)
        self.choices.remove(choice)
        self._ensure_choice_positions()
        choice.delete()
        return choice

    def move_choice(self, choice_label, new_index):
        choice = self.get_choice(choice_label)
        old_index = self.choices.index(choice)
        self.choices.insert(new_index, self.choices.pop(old_index))
        self._ensure_choice_positions()
        return choice

class Choice(models.Model):
    field    = models.ForeignKey(Field, related_name="_choice_set")
    slug     = models.SlugField(editable=False)
    label    = models.CharField(max_length=250)
    position = models.IntegerField()

    def save(self, *args, **kwargs):
        self.slug = slugify(self.label).replace("-", "_")[:50]
        super(Choice, self).save(*args, **kwargs)



from django import forms
from django.core.urlresolvers import reverse
from django.test import TestCase
from django.utils import simplejson as json

from .exceptions import (ChoiceDoesNotExist, ChoiceAlreadyExists,
                         FieldDoesNotExist, FieldAlreadyExists)
from .models import Form, Field, Choice
from .transactions import Transaction

class BaseTestCase(TestCase):
    def setUp(self):
        self.form = Form.objects.create(name="Test Form",
                                        description="A test form")

    def get(self, url_name, *args, **kwargs):
        return self.client.get(reverse(url_name, args=args, kwargs=kwargs))

    def post(self, url_name, *args, **kwargs):
        data = kwargs.pop("data", None)
        extra = kwargs.pop("extra", {})
        return self.client.post(reverse(url_name, args=args, kwargs=kwargs),
                                data,
                                **extra)

class AddRemoveFieldTestCase(BaseTestCase):
    def test_add_field(self):
        self.assertEqual(len(self.form.fields), 0,
                         "Should start with 0 fields")
        self.form.add_field("Email", type="EmailField")
        self.assertEqual(len(self.form.fields), 1,
                         "Should now have 1 field")
        self.assertTrue(isinstance(self.form.fields[0], Field),
                        "The new field should be an instance of Field.")

    def test_add_existing_field(self):
        self.form.add_field("Email")
        try:
            self.form.add_field("Email")
        except FieldAlreadyExists:
            self.assertTrue(True,
                            "Adding a field that already exists should throw an error.")
        else:
            self.assertTrue(False,
                            "Adding a field that already exists should throw an error.")

    def test_remove_field(self):
        self.form.add_field("Email", type="EmailField")
        self.assertEqual(len(self.form.fields), 1,
                         "Should have 1 field")
        self.form.remove_field("Email")
        self.assertEqual(len(self.form.fields), 0,
                         "Should now have 0 fields")

    def test_remove_nonexistant_field(self):
        try:
            self.form.remove_field("non-existant field")
        except FieldDoesNotExist:
            self.assertTrue(True,
                            "Removing non-existant field throws an error")
        else:
            self.assertTrue(False,
                            "Removing non-existant field throws an error")

class AddRemoveChoiceTestCase(BaseTestCase):
    def setUp(self):
        super(AddRemoveChoiceTestCase, self).setUp()
        self.field = self.form.add_field("Happy and you know it?",
                                         type="MultipleChoiceField")

    def test_add_choice(self):
        self.assertEqual(len(self.field.choices), 0,
                         "Should start with 0 choices")
        self.field.add_choice("Yes")
        self.assertEqual(len(self.field.choices), 1,
                         "Should now have 1 choice")
        self.assertTrue(isinstance(self.field.choices[0], Choice),
                        "The new choice should be an instance of Choice.")

    def test_add_existing_choice(self):
        self.field.add_choice("Yes")
        try:
            self.field.add_choice("Yes")
        except ChoiceAlreadyExists:
            self.assertTrue(True,
                            "Adding a choice that already exists should throw an error.")
        else:
            self.assertTrue(False,
                            "Adding a choice that already exists should throw an error.")

    def test_remove_choice(self):
        self.field.add_choice("No")
        self.assertEqual(len(self.field.choices), 1,
                         "Should have 1 choice")
        self.field.remove_choice("No")
        self.assertEqual(len(self.field.choices), 0,
                         "Should now have 0 fields")

    def test_remove_nonexistant_choice(self):
        try:
            self.field.remove_choice("non-existant choice")
        except ChoiceDoesNotExist:
            self.assertTrue(True,
                            "Removing non-existant choice throws an error")
        else:
            self.assertTrue(False,
                            "Removing non-existant choice throws an error")

class ModelToDjangoFormTestCase(BaseTestCase):
    def setUp(self):
        super(ModelToDjangoFormTestCase, self).setUp()
        self.field = self.form.add_field("Happy and you know it?",
                                         type="MultipleChoiceField")
        self.field.add_choice("yes")
        self.field.add_choice("no")

    def test_create_django_form_field(self):
        form_field = self.field.as_django_form_field()
        self.assertTrue(isinstance(form_field, forms.fields.Field),
                        "field.as_django_form_field() should return an instance of a django Field.")
        self.assertEqual(len(form_field.choices), len(self.field.choices),
                         "The django form field should have same number of choices as field model instance.")

    def test_create_django_form(self):
        TestForm = self.form.as_django_form()
        self.assertTrue(issubclass(TestForm, forms.Form),
                        "form.as_django_form() should return a subclass of a django Form class")
        self.assertEqual(len(TestForm().fields), len(self.form.fields),
                         "The django form should have same number of fields as the form model instance.")

class ApplyTransactionsViewTestCase(BaseTestCase):
    def test_apply_transactions(self):
        transactions = json.dumps([{ "action"       : "add field",
                                     "label"        : "hello" },
                                   { "action"       : "change field type",
                                     "label"        : "hello",
                                     "to"           : "MultipleChoiceField" },
                                   { "action"       : "add choice",
                                     "label"        : "hello",
                                     "choice_label" : "uno" },
                                   { "action"       : "add choice",
                                     "label"        : "hello",
                                     "choice_label" : "dos" },
                                   { "action"       : "add field",
                                     "label"        : "hola" },
                                   { "action"       : "change name",
                                     "to"           : "my form"}])

        response = self.post("wysiwyg_forms_apply_transactions",
                             self.form.id,
                             data={ "transactions" : transactions })
        self.assertEqual(response.status_code, 200)

        # Get the latest form instance
        self.form = Form.objects.get(id=self.form.id)

        self.assertEqual(len(self.form.fields), 2)
        self.assertEqual(len(self.form.get_field("hello").choices), 2)

        # TODO: Test the response JSON

class NewFormViewTestCase(BaseTestCase):
    def test_a_new_form_is_created(self):
        num_forms = Form.objects.all().count()
        response = self.get("wysiwyg_forms_new_form")
        self.assertEqual(response.status_code, 200)
        self.assertEqual(num_forms + 1, Form.objects.all().count())

class TransactionsTestCase(BaseTestCase):
    def setUp(self):
        super(TransactionsTestCase, self).setUp()
        self.field = self.form.add_field("Happy and you know it?",
                                         type="MultipleChoiceField")
        self.field.add_choice("yes")
        self.field.add_choice("no")

    def test_change_name(self):
        transaction = Transaction(action="change name",
                                  to="New Name")
        transaction.apply_to(self.form)
        self.assertEqual(self.form.name, "New Name")

    def test_change_description(self):
        transaction = Transaction(action="change description",
                                  to="New description")
        transaction.apply_to(self.form)
        self.assertEqual(self.form.description, "New description")

    def test_add_field(self):
        num_fields = len(self.form.fields)
        transaction = Transaction(action="add field",
                                  label="some new field")
        transaction.apply_to(self.form)
        self.assertEqual(num_fields + 1, len(self.form.fields))
        self.assertTrue(any(f.label == "some new field"
                            for f in self.form.fields))

    def test_remove_field(self):
        num_fields = len(self.form.fields)
        transaction = Transaction(action="remove field",
                                  label="Happy and you know it?")
        transaction.apply_to(self.form)
        self.assertEqual(num_fields - 1, len(self.form.fields))
        self.assertTrue(not any(f.label == "Happy and you know it?"
                                for f in self.form.fields))

    def test_rename_field(self):
        transaction = Transaction(action="rename field",
                                  label="Happy and you know it?",
                                  to="new label")
        transaction.apply_to(self.form)
        self.assertTrue(any(f.label == "new label"
                            for f in self.form.fields))
        self.assertTrue(not any(f.label == "Happy and you know it?"
                                for f in self.form.fields))

    def test_change_help_text(self):
        transaction = Transaction(action="change help text",
                                  label="Happy and you know it?",
                                  to="helpful")
        transaction.apply_to(self.form)
        self.assertEqual(self.field.help_text, "helpful")

    def test_move_field(self):
        self.form.add_field("a field")
        self.form.add_field("another field")
        self.form.add_field("some other field")
        num_fields = len(self.form.fields)
        transaction = Transaction(action="move field",
                                  label="a field",
                                  to=1)
        transaction.apply_to(self.form)
        self.assertEqual(self.form.fields[1].label, "a field")
        self.assertEqual(num_fields, len(self.form.fields))

    def test_change_field_type(self):
        transaction = Transaction(action="change field type",
                                  label="Happy and you know it?",
                                  to="BooleanField")
        transaction.apply_to(self.form)
        self.assertTrue(isinstance(self.field.as_django_form_field(),
                                   forms.fields.BooleanField))

    def test_change_field_widget(self):
        transaction = Transaction(action="change field widget",
                                  label="Happy and you know it?",
                                  to="RadioSelect")
        transaction.apply_to(self.form)
        self.assertTrue(isinstance(self.form.get_field("Happy and you know it?").as_django_form_field().widget,
                                   forms.widgets.RadioSelect))

    def test_change_field_required(self):
        self.assertEqual(self.form.get_field("Happy and you know it?").as_django_form_field().required,
                         True)
        transaction = Transaction(action="change field required",
                                  label="Happy and you know it?",
                                  to=False)
        transaction.apply_to(self.form)
        self.assertEqual(self.form.get_field("Happy and you know it?").as_django_form_field().required,
                         False)

    def test_add_choice(self):
        num_choices = len(self.field.choices)
        transaction = Transaction(action="add choice",
                                  label="Happy and you know it?",
                                  choice_label="possibly")
        transaction.apply_to(self.form)
        self.assertEqual(num_choices + 1, len(self.field.choices))
        self.assertTrue(any(c.label == "possibly"
                            for c in self.field.choices))

    def test_remove_choice(self):
        num_choices = len(self.field.choices)
        transaction = Transaction(action="remove choice",
                                  label="Happy and you know it?",
                                  choice_label="no")
        transaction.apply_to(self.form)
        self.assertEqual(num_choices - 1, len(self.field.choices))
        self.assertTrue(not any(c.label == "no"
                                for c in self.field.choices))

    def test_change_choice(self):
        transaction = Transaction(action="change choice",
                                  label="Happy and you know it?",
                                  choice_label="yes",
                                  to="Hell yeah!")
        transaction.apply_to(self.form)
        self.assertTrue(not any(c.label == "yes"
                                for c in self.field.choices))
        self.assertTrue(any(c.label == "Hell yeah!"
                            for c in self.field.choices))

    def test_move_choice(self):
        self.field.add_choice("maybe")
        self.field.add_choice("sure")
        self.field.add_choice("eh eh")
        num_choices = len(self.field.choices)
        transaction = Transaction(action="move choice",
                                  label="Happy and you know it?",
                                  choice_label="sure",
                                  to=0)
        transaction.apply_to(self.form)
        self.assertEqual(num_choices, len(self.field.choices))
        self.assertEqual(self.field.choices[0].label, "sure")



class WysiwygFormsException(Exception):
    pass

class ChoiceDoesNotExist(WysiwygFormsException):
    pass

class ChoiceAlreadyExists(WysiwygFormsException):
    pass

class FieldDoesNotExist(WysiwygFormsException):
    pass

class FieldAlreadyExists(WysiwygFormsException):
    pass



from django import forms

def is_valid_field_type(field_type):
    field = getattr(forms.fields, field_type, False)
    return field and issubclass(field, forms.fields.Field)

def is_valid_widget_type(widget_type):
    widget = getattr(forms.widgets, widget_type, False)
    return widget and issubclass(widget, forms.widgets.Widget)

def field_type_has_choices(field_type):
    field = getattr(forms.fields, field_type, False)
    if not field:
        return False
    else:
        return hasattr(field(), "choices")



from .exceptions import WysiwygFormsException
from .utils import field_type_has_choices, is_valid_field_type
from .models import Choice

__all__ = ("Transaction",)

class Transaction(object):
    """
    A transaction represents an atomic change that can be applied to a form with
    `apply_to`.
    """

    # This dictionary is keyed by action and has values which are functions that
    # take a transaction whose action is the same as the key and a form to apply
    # the transaction to. To add a function to this dictionary, use the
    # `@Transaction.register(some_action)` decorator.
    _action_methods = {}

    def __init__(self, **kwargs):
        self.action       = unicode(kwargs.pop("action", kwargs.pop(u"action", None)))
        self.to           = kwargs.pop("to", kwargs.pop(u"to", None))
        self.label        = kwargs.pop("label", kwargs.pop(u"label", None))
        self.choice_label = kwargs.pop("choice_label", kwargs.pop(u"choice_label", None))

        try:
            key = iter(kwargs).next()
            raise WysiwygFormsException(
                "Transaction.__init__() got unexpected keyword argument '%s'" % key
                )
        except StopIteration:
            pass

    def __unicode__(self):
        return u"<%s: %s>" % (self.__class__.__name__, self.__dict__)
    __str__ = __repr__ = __unicode__

    @classmethod
    def register(cls, action):
        """
        This decorator defines new types of Transaction actions, and specifies
        how to apply the new type of action to a form.
        """
        def registrar(fn):
            cls._action_methods[unicode(action)] = fn
            return fn
        return registrar

    def apply_to(self, form):
        """
        Apply this transaction to the given form.
        """
        fn = Transaction._action_methods.get(self.action)
        if not fn:
            raise WysiwygFormsException("Unknown action to apply to form: '%s'" % self.action)
        else:
            fn(self, form)


@Transaction.register("change name")
def apply_to(t, form):
    form.name = t.to

@Transaction.register("change description")
def apply_to(t, form):
    form.description = t.to

@Transaction.register("add field")
def apply_to(t, form):
    form.add_field(t.label)

@Transaction.register("remove field")
def apply_to(t, form):
    form.remove_field(t.label)

@Transaction.register("rename field")
def apply_to(t, form):
    form.get_field(t.label).label = t.to

@Transaction.register("change help text")
def apply_to(t, form):
    form.get_field(t.label).help_text = t.to

@Transaction.register("move field")
def apply_to(t, form):
    form.move_field(t.label, t.to)

@Transaction.register("change field type")
def apply_to(t, form):
    if not is_valid_field_type(t.to):
        raise WysiwygFormsException("Invalid field type: %s" % t.to)
    else:
        field = form.get_field(t.label)
        field.type = t.to
        if not field_type_has_choices(t.to):
            Choice.objects.filter(field=field).delete()

@Transaction.register("change field widget")
def apply_to(t, form):
    form.get_field(t.label).widget = t.to

@Transaction.register("change field required")
def apply_to(t, form):
    form.get_field(t.label).required = t.to

@Transaction.register("add choice")
def apply_to(t, form):
    form.get_field(t.label).add_choice(t.choice_label)

@Transaction.register("remove choice")
def apply_to(t, form):
    form.get_field(t.label).remove_choice(t.choice_label)

@Transaction.register("change choice")
def apply_to(t, form):
    form.get_field(t.label).get_choice(t.choice_label).label = t.to

@Transaction.register("move choice")
def apply_to(t, form):
    form.get_field(t.label).move_choice(t.choice_label, t.to)



from django import http
from django import template
from django.conf import settings
from django.shortcuts import get_object_or_404
from django.template import RequestContext
from django.utils import simplejson as json
from django.views.generic import DetailView

from .exceptions import WysiwygFormsException
from .models import Form
from .transactions import Transaction

__all__ = ("ApplyTransactions", "Edit")

class ApplyTransactions(DetailView):
    queryset = Form.objects.all()
    context_object_name = "form"
    http_method_names = ["post"]

    def __init__(self, *args, **kwargs):
        super(ApplyTransactions, self).__init__(*args, **kwargs)
        self.error = None
        self.post = self.get

    def render_to_response(self, context):
        return http.HttpResponse(json.dumps(context, cls=Form.JSONEncoder),
                                 content_type="application/json",
                                 status=self.get_status())

    def get_object(self, **kwargs):
        form = super(ApplyTransactions, self).get_object(**kwargs)
        try:
            for t in json.loads(self.request.POST.get("transactions", "[]")):
                Transaction(**t).apply_to(form)
            form.save()
        except WysiwygFormsException, e:
            self.error = e
        return form

    def get_context_data(self, **kwargs):
        context = super(ApplyTransactions, self).get_context_data(**kwargs)
        if self.error:
            return { "error": self.error.message }
        else:
            return context["form"]

    def get_status(self):
        if self.error:
            return 500
        else:
            return 200

class Edit(DetailView):
    template_name = "wysiwyg_forms/edit.html"
    queryset = Form.objects.all()
    context_object_name = "form"

    # Customize `base_template_name` to change what template
    # `wysiwyg_forms/edit.html` will extend.
    base_template_name = "wysiwyg_forms/base.html"

    def get_object(self, queryset=None):
        try:
            form = super(Edit, self).get_object(queryset)
        except AttributeError:
            form = Form.objects.create(name="New Form",
                                       description="This is a new form.")
        return form

    def get_context_data(self, **kwargs):
        context = super(Edit, self).get_context_data(**kwargs)
        context["base_template_name"] = self.base_template_name
        context["debug"] = settings.DEBUG
        return context



from django.conf import settings
from django.conf.urls.defaults import *

from django.views.generic import TemplateView

from .views import ApplyTransactions, Edit

urlpatterns = patterns(
    '',
    url(r"^save/(?P<pk>\d+)/$",
        ApplyTransactions.as_view(),
        name="wysiwyg_forms_apply_transactions"),
    url(r"^$",
        Edit.as_view(),
        name="wysiwyg_forms_new_form"),
    url(r"^(?P<pk>\d+)/$",
        Edit.as_view(),
        name="wysiwyg_forms_edit_form")
)

if settings.DEBUG:
    urlpatterns += patterns(
        '',
        url(r"^test/$",
            TemplateView.as_view(template_name="wysiwyg_forms/test.html"),
            name="wysiwyg_forms_test")
    )



from setuptools import setup, find_packages

setup(name = "django-wysiwyg-forms",
      version = "0.1.0",
      description = "A What You See Is What You Get form editor for Django.",
      author = "Nick Fitzgerald",
      author_email = "fitzgen@gmail.com",
      url = "https://github.com/fitzgen/django-wysiwyg-forms",
      packages = find_packages(),
      install_requires = ['django==1.3'])
